from typing import List

import jax
import jax.numpy as jnp


def cast_arrays(array: List[jax.Array], dtype):
    if dtype is not None:
        return [jnp.astype(el, dtype) for el in array]
    else:
        return array


def check_shape(*args):
    if not args:
        return False

    shape = args[0].shape

    for arg in args[1:]:
        if arg.shape != shape:
            return False

    return True



import numpy as np
from scipy.stats import poisson


class MertonSimulator:

    """
    Simulation of the underlying according to Merton model
    -----
    S0:     price at time 0 [float]
    T:      maturity in yearfrac [float]
    r:      risk free rate [float]
    N_SIM:  number of simulation [int]
    N:      number of the monitoring data [int]
    SIGMA:  underlying volatility [float]
    PARAMS: vector of merton parameters
    """

    def __init__(self, S0, T, r, N_SIM, N, PARAMS):
        self.S0 = S0
        self.T = T
        self.r = r
        self.N_SIM = N_SIM
        self.N = N
        self.PARAMS = PARAMS

    def simulate_merton(self):
        sigma, lambda_mert, mu, delta = self.PARAMS

        # delta time
        dt = self.T / self.N

        # initialize X
        x = np.zeros((self.N_SIM, self.N + 1))  # X = rt + X(t)

        # number of jumps
        NT = poisson.ppf(np.random.rand(self.N_SIM, 1), lambda_mert * self.T)

        # characteristic exponent
        charexp = lambda u: -0.5 * sigma ** 2 * u ** 2 + lambda_mert * (np.exp(-0.5 * delta ** 2 * u ** 2 + 1j * mu * u) - 1)
        drift = self.r - charexp(-1j)  # in order to be risk-neutral

        for j in range(self.N_SIM):
            JumpTimes = np.sort(self.T * np.random.rand(int(NT[j]), 1))

            for i in range(self.N):
                # add diffusion component
                x[j, i + 1] = x[j, i] + drift * dt + sigma * np.sqrt(dt) * np.random.randn()

                # add jump part if exists jump in ((i-1)*dt, i*dt)
                for l in range(int(NT[j])):
                    if (i - 1) * dt < JumpTimes[l] <= i * dt:
                        Y = mu + delta * np.random.randn()
                        x[j, i + 1] = x[j, i + 1] + Y

        S = self.S0 * np.exp(x)

        return S


"""
            # Example usage:
            params = [0.2, 0.1, 0.02, 0.1]  # Replace with the actual values
            merton_simulator = MertonSimulator(100, 1, 0.05, 1000, 252, params)
            result = merton_simulator.simulate_merton()
            print(result)
"""


import jax
import jax.numpy as jnp
from jax.scipy.special import erf
from jax.scipy.stats.norm import pdf

_SQRT_2 = jnp.sqrt(2.0)


@jax.jit
def d1(spots, strikes, vols, expires, discount_rates):
    vol_sqrt_t = vols * jnp.sqrt(expires)

    return jnp.divide((jnp.log(spots / strikes) + (discount_rates + (vols ** 2 / 2)) * expires), vol_sqrt_t)


@jax.jit
def cum_normal(x):
    return (erf(x / _SQRT_2) + 1) / 2


@jax.jit
def density_normal(x):
    return pdf(x)



"""Option prices computed with the Black-Scholes module"""
from jaxfin.price_engine.black_scholes.vanilla_options import delta_vanilla, bs_price, gamma_vanilla

european_price = bs_price
delta_european = delta_vanilla
gamma_european = gamma_vanilla


from .math import *


def _compute_d1_d2(spots, strikes, expires, vols, discount_rates):
    vol_sqrt_t = vols * jnp.sqrt(expires)

    _d1 = d1(spots, strikes, vols, expires, discount_rates)

    return [_d1, _d1 - vol_sqrt_t]


@jax.jit
def compute_undiscounted_call_prices(spots, strikes, expires, vols, discount_rates):
    [_d1, _d2] = _compute_d1_d2(spots, strikes, expires, vols, discount_rates)

    return cum_normal(_d1) * spots - cum_normal(_d2) * strikes


@jax.jit
def compute_discounted_call_prices(spots, strikes, expires, vols, discount_rates):
    [_d1, _d2] = _compute_d1_d2(spots, strikes, expires, vols, discount_rates)

    return cum_normal(_d1) * spots - cum_normal(_d2) * strikes * jnp.exp((- discount_rates) * expires)



import numpy as np
from scipy.stats import poisson, expon

class KouSimulator:
    def __init__(self, SIGMA, LAMBDA, LAMBDA_MINUS, LAMBDA_PLUS, P):
        self.SIGMA = SIGMA
        self.LAMBDA = LAMBDA
        self.LAMBDA_MINUS = LAMBDA_MINUS
        self.LAMBDA_PLUS = LAMBDA_PLUS
        self.P = P

    def charexp(self, u):
        return -self.SIGMA**2 * u**2 / 2 + 1j * u * self.LAMBDA * (
            self.P / (self.LAMBDA_PLUS - 1j * u) - (1 - self.P) / (self.LAMBDA_MINUS + 1j * u)
        )

    def simulate_kou(self, S0, T, r, N_SIM, N):
        # delta time
        dt = T / N

        # initialize X
        X = np.zeros((N_SIM, N + 1))  # X = rt + X(t)

        # number of jumps
        NT = poisson.ppf(np.random.rand(N_SIM, 1), self.LAMBDA * T)

        # characteristic exponent
        drift = r - self.charexp(-1j)  # in order to be risk-neutral

        for j in range(N_SIM):
            JumpTimes = np.sort(T * np.random.rand(int(NT[j]), 1))

            for i in range(N):
                # add diffusion component
                X[j, i + 1] = X[j, i] + drift * dt + self.SIGMA * np.sqrt(dt) * np.random.randn()

                # add jump part -> ( (i-1)dt, idt )
                for l in range(int(NT[j])):
                    if (i - 1) * dt < JumpTimes[l] <= i * dt:
                        sim_p = np.random.rand()

                        if sim_p < self.P:  # positive jump
                            Y = expon.ppf(np.random.rand(), scale=1 / self.LAMBDA_PLUS)
                        else:  # negative jump
                            Y = -expon.ppf(np.random.rand(), scale=1 / self.LAMBDA_MINUS)

                        X[j, i + 1] = X[j, i + 1] + Y

        S = S0 * np.exp(X)

        return S

"""
# Example usage:
kou_simulator = KouSimulator(0.2, 0.1, 0.05, 0.15, 0.8)  # Replace with the actual values
result = kou_simulator.simulate_kou(100, 1, 0.05, 1000, 252)
print(result)
"""